# -*- coding: utf-8 -*-

"""Module containing the core functionality."""
from __future__ import unicode_literals  # unicode support for py2

import collections
import glob
import re

import click

from .client import WoppClient
from .constants import (
    REQUIREMENTS_REPLACE_COMMENT,
    REQ_LINE_REGEX,
    ALL_OPTION,
)
from .exceptions import (
    DocsNotFoundError,
    URLLaunchError,
    RequirementsFilesNotFoundError
)
from .param_types import MultipleChoice
from .utils import clean_response, extract_pkg_version


def get_output(response, more_out=False):
    """
    Returns final output to display

    :param response: WoppResponse object to get the info
    :param more_out: more or less information?
    :return: dict containing output information
    """
    out_dict = {
        'name': response.name,
        'current_version': response.latest_version,
        'summary': response.summary,
        'homepage': response.homepage,
        'package_url': response.package_url,
        'author': response.author,
    }

    if more_out:
        out_dict.update({
            'author_email': response.author_email,
            'releases': ', '.join(response.releases) if response.releases else None,
            'project_urls': response.project_urls,
            'requires_python': response.requires_python,
            'license': response.license,
            'current_release_url': response.latest_release_url,
            'current_package_info': response.latest_pkg_urls,
            'dependencies': ', '.join(response.dependencies) if response.dependencies else None,
        })
    else:
        out_dict.update({
            'latest_releases': ', '.join(response.latest_releases) if response.latest_releases else None,
        })

    return out_dict


def get_req_files(req_dir, req_pattern):
    req_files = glob.glob(req_dir + "/" + req_pattern)
    # prompt user only if there's more than one requirements file.
    num_req_files = len(req_files)

    if not num_req_files:
        raise RequirementsFilesNotFoundError(
            "No files were found matching pattern '{}' in the provided directory path :\n{}".format(
                req_pattern, req_dir
            ))

    # if there's only one file available, don't prompt.
    if num_req_files > 1:
        file_choice_list = [ALL_OPTION]
        file_choice_list.extend(req_files)
        # add the `all` option

        choice_map = collections.OrderedDict(
            ('{}'.format(i), value) for i, value in enumerate(file_choice_list, 1)
        )  # creates dict of form {'1': './requirements.txt', ...}

        choices = choice_map.keys()
        default_choice = '1'

        # build prompt
        choice_lines = ['{} - {}'.format(k, v) for k, v in choice_map.items()]
        prompt_text = '\n'.join([
            "We found {} files matching the pattern '{}'. Please choose if"
            " you'd like to modify just one, many or all of them. You can specify"
            " one or multiple options with a comma.. \nExamples:\n1,2,3\n1\n"
            "'{}' means that all files will be checked and modified, which is the"
            " default. Hit Ctrl+C to quit.".format(num_req_files, req_pattern, ALL_OPTION),
            '\n'.join(choice_lines),
            "Hit Enter to use the default or choose from the options",
        ])

        # prompt user
        req_file_ids = click.prompt(
            prompt_text,
            type=MultipleChoice(choices),
            default=default_choice,
            show_default=True,
            show_choices=True,
        )
        # if the all option is not included in the input,
        # only use the files that are needed.
        if default_choice not in req_file_ids:
            req_files = [choice_map[str(id_)] for id_ in req_file_ids if id_ != default_choice]

    return req_files


def add_pkg_to_req(package, version, req_dir, req_pattern, comment):
    req_files = get_req_files(req_dir, req_pattern)
    req_line = "{}=={}".format(package, version)

    repl_str = ''
    if comment:
        # add comment if provided.
        repl_str += '# {}\n'.format(comment)
    repl_str += req_line

    click.echo("Adding {} ...".format(req_line))
    for file_path in req_files:
        needs_append = True  # should we append package to the end of file

        click.echo("Modifying file: {} ...".format(file_path))
        with open(file_path, 'r+') as file:
            # read all lines at once into memory.
            # NOTE: This is not memory efficient, but requirements files are small
            # so this is good for now. Another way would be to create another file
            # and write to it, but that seems overkill for requirements files.
            data = file.readlines()
            for line_num, line in enumerate(data):
                line = line.strip().lower()
                if line:
                    # first, search if the pkg already exists
                    if not line.startswith("#"):  # not a comment
                        package_, version_ = extract_pkg_version(line)

                        if package.lower() == package_:
                            needs_append = False
                            # if yes, check the version.
                            if version != version_:
                                # replace and end it.
                                data[line_num] = re.sub(REQ_LINE_REGEX, req_line, line)
                            else:
                                click.echo("Package is already set to the latest/desired version.")
                            break

                    # if pkg is absent, check for the '#wopp' comment.
                    # handle empty spaces
                    if REQUIREMENTS_REPLACE_COMMENT.lower() in line.replace(' ', ''):
                        needs_append = False

                        data[line_num] = line.replace(line, repl_str + "\n")
                        # only replace the first instance of #wopp
                        break

            # move pointer back to start
            file.seek(0, 0)
            # write data
            file.writelines(data)
            # truncate rest of the file
            file.truncate()

        if needs_append:
            # if none of the above cases happen,
            # just append to the end of the file and done.
            with open(file_path, 'a') as file:
                file.write("\n{}\n".format(repl_str))


def run_query(
    package=None,
    version=None,
    more_out=False,
    launch_docs=False,
    add_to_req=False,
    req_dir=None,
    req_pattern=None,
    comment=None,
):
    """
    Run query against PyPI API and then do stuff based on user options.

    :param package: name of package
    :param version: version of package
    :param more_out: should output should contain more detail?
    :param launch_docs: should doc URL be launched?
    :param add_to_req: should the package be added as a dependency to requirements files?
    :param comment: comment to be added for the dependency
    :param req_dir: Directory to search for requirement files
    :param req_pattern: Filename pattern for searching requirements files
    :return: output if available, or None
    """
    client = WoppClient(request_hooks={'response': clean_response})
    response = client.request(package=package, version=version)

    # launch of docs url
    if launch_docs:
        url = response.project_docs
        if not url:
            raise DocsNotFoundError("Could not find any documentation or homepage URL to launch.")

        exit_status = click.launch(url)
        if exit_status:  # if 1
            raise URLLaunchError("There was a problem opening the URL in your browser.")

        return

    # add pkg as dep to requirements files if needed.
    if add_to_req:
        add_pkg_to_req(response.name, version or response.latest_version, req_dir, req_pattern, comment)
        return

    # get the output
    out_dict = get_output(response, more_out=more_out)
    # default out
    return out_dict



# -*- coding: utf-8 -*-

"""
Package level constants
"""
from __future__ import unicode_literals  # unicode support for py2

PYPI_BASE_URL = 'https://pypi.org/pypi'

REQUIREMENTS_REPLACE_COMMENT = "#wopp"
ALL_OPTION = 'ALL'
REQ_LINE_REGEX = "([A-Za-z0-9_-]+)((==)|(>=)|(<=))([A-Za-z0-9_.]+)"



# -*- coding: utf-8 -*-

"""Utility methods"""
from __future__ import unicode_literals  # unicode support for py2

import re

import click

from .constants import REQ_LINE_REGEX


def extract_pkg_version(in_str):
    """
    Use regex to extract package and version

    :param in_str: input string
    :return: tuple of pkg, version
    """
    package = None
    version = None

    try:
        reg_search = re.search(REQ_LINE_REGEX, in_str)
        if reg_search:
            package = reg_search.group(1)
            version = reg_search.group(6)
    except IndexError:
        pass

    return package, version


def pretty(input_, indent=0):
    """
    Pretty print dictionary

    :param input_: input
    :param indent: number of tabs
    :return: None
    """

    def get_readable_key(key_):
        # capitalize and remove _
        if '_' in key_:
            return key_.upper().replace('_', ' ')
        else:
            return key_.upper()

    if isinstance(input_, dict):
        for key, value in input_.items():
            # only print if there's something
            if value:
                click.secho('\t' * indent + get_readable_key(str(key)), fg='green', bold=True)
                if isinstance(value, dict):
                    pretty(value, indent + 1)
                else:
                    click.echo('\t' * (indent + 1) + str(value))
    else:
        click.echo(input_)


def clean_response(r, *args, **kwargs):
    """
    Hook called after a response is received.
    Used to modify response.

    :param r: requests.models.Response object
    :param args:
    :param kwargs:
    :return: modified Response object
    """

    def convert_pkg_info(pkg_url_list):
        """
        Converts a list of package info dicts
        into a dict, where the key is the type
        of package.. eg: sdist
        :param pkg_url_list:
        :return: dict
        """
        package_urls = {}
        for pkg_url in pkg_url_list:
            package_urls.update({
                pkg_url.get('packagetype'): {
                    'md5': pkg_url.get('digests').get('md5'),
                    'sha256': pkg_url.get('digests').get('sha256'),
                    'filename': pkg_url.get('filename'),
                    'has_sig': pkg_url.get('has_sig'),
                    'size': pkg_url.get('size'),
                    'upload_time': pkg_url.get('upload_time'),
                    'url': pkg_url.get('url'),
                }})
        return package_urls

    # only run hooks for 200
    if r.status_code == 200:
        r.hook_called = True  # flag to confirm hook was called.

        dirty_response = r.json()
        cleaned_response = {}

        info = dirty_response.get('info')
        if info:
            cleaned_response = {
                'name': info.get('name'),
                'latest_version': info.get('version'),
                'summary': info.get('summary'),
                'homepage': info.get('home_page'),
                'package_url': info.get('project_url') or info.get('package_url'),
                'author': info.get('author'),
                'project_urls': info.get('project_urls'),
                'requires_python': info.get('requires_python'),
                'license': info.get('license'),
                'author_email': info.get('author_email'),
                'latest_release_url': info.get('release_url'),
                'dependencies': info.get('requires_dist'),
            }

        # release list
        releases = dirty_response.get('releases')
        if releases:
            release_list = list(releases.keys())
            release_list.reverse()

            # more detailed info of every release's package
            releases_info = {}
            for key, val in releases.items():
                if val:
                    releases_info[key] = convert_pkg_info(val)

            cleaned_response.update({
                'releases': release_list,
                'releases_pkg_info': releases_info,
            })

        # latest release's package information
        latest_pkg_urls = dirty_response.get('urls')
        if latest_pkg_urls:
            cleaned_response.update({
                'latest_pkg_urls': convert_pkg_info(latest_pkg_urls),
            })

        r.cleaned_json = cleaned_response
    return r



# -*- coding: utf-8 -*-

from __future__ import unicode_literals  # unicode support for py2

import click


class MultipleChoice(click.Choice):
    """
    extension of click.Choice that
    accepts multiple choice inputs
    and converts them into a list
    """
    name = 'Multiple Choice Param Type'

    def convert(self, value, param, ctx):
        cleaned_value = value.strip()

        if cleaned_value in self.choices:
            return [cleaned_value]

        choice_list = cleaned_value.split(',')
        valid_choice_list = []

        for choice in choice_list:
            choice = choice.strip().lower()  # lower is for when we have a,b,c as options.
            if choice not in self.choices:
                self.fail('Invalid choice: %s (choose from %s)' %
                          (choice, ', '.join(self.choices)), param, ctx)

            valid_choice_list.append(choice)

        return valid_choice_list



# -*- coding: utf-8 -*-

"""
Console script
"""
from __future__ import unicode_literals  # unicode support for py2

import click

from whatsonpypi import __version__
from .utils import pretty, extract_pkg_version
from .whatsonpypi import run_query


@click.command(context_settings=dict(help_option_names=['-h', '--help']))
@click.version_option(__version__, '-v', '--version')
@click.argument('package')
@click.option(
    '-m',
    '--more',
    is_flag=True,
    required=True,
    default=False,
    show_default=True,
    help="Flag to enable expanded output"
)
@click.option(
    '-d',
    '--docs',
    is_flag=True,
    required=False,
    default=False,
    help="Flag to open docs or homepage of project"
)
@click.option(
    '-a',
    '--add',
    is_flag=True,
    required=False,
    default=False,
    help="Flag to enable adding of dependencies to requirement files."
         " By default, it searches for files with names matching requirements*.txt"
         " in the current working directory and adds the dependency to the end of the"
         " file. If you want the dependency to be added to a specific line,"
         " mention a comment '#wopp' on its own line which will be replaced with the dependency."
         " Existing dependencies will be replaced with newer versions. Dependency version"
         " by default is the latest unless specified explicitly with 'whatsonpypi package==version'."
         " Directory to search for requirement files can be specified with --req-dir"
)
@click.option(
    '-r',
    '--req-dir',
    type=click.Path(exists=True, file_okay=False, dir_okay=True,
                    readable=True, writable=True, resolve_path=True,
                    allow_dash=False),
    required=False,
    default=".",
    show_default=True,
    help="Directory to search for requirement files. Only used when --add is used."
)
@click.option(
    '-p',
    '--req-pattern',
    type=str,
    required=True,
    default="requirements*.txt",
    show_default=True,
    help="Filename pattern for searching requirements files."
)
@click.option(
    '-c',
    '--comment',
    type=str,
    required=False,
    show_default=False,
    help="Comment to be added for the dependency when using --add."
)
def main(package, more, docs, add, req_dir, req_pattern, comment):
    """
    CLI tool to get package info from PyPI and/or manipulate requirements.

    Example usages:

    $ whatsonpypi django
    """
    try:
        # get version if given
        package_, version = extract_pkg_version(package)

        result = run_query(
            package=package_ or package,
            version=version,
            more_out=more,
            launch_docs=docs,
            add_to_req=add,
            req_dir=req_dir,
            req_pattern=req_pattern,
            comment=comment,
        )
        # output is not always expected and might be None sometimes.
        if result:
            pretty(result)
    except Exception as e:
        # all other exceptions
        raise click.ClickException(e)


if __name__ == "__main__":
    main()  # pragma: no cover



# -*- coding: utf-8 -*-

"""
CLI entry point
"""
# enable absolute imports of this module for Python2.x
from __future__ import absolute_import

from .cli import main

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

"""
whatsonpypi.exceptions
-----------------------
All exceptions used in the whatsonpypi code base are defined here.
"""


class WoppException(Exception):
    """
    Base exception. All other exceptions
    inherit from here.
    """


class PackageNotProvidedError(WoppException):
    """
    Raised when no package is available for the client to request
    """


class PackageNotFoundError(WoppException):
    """
    Raised when a package is not found on PyPI
    """


class DocsNotFoundError(WoppException):
    """
    Raised when a package does not have documentation or homepage urls
    """


class URLLaunchError(WoppException):
    """
    Raised when there's a problem opening a URL in the browser
    """


class RequirementsFilesNotFoundError(WoppException):
    """
    Raised when no requirements files are found in provided directory path.
    """



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""
import io

from setuptools import setup, find_packages

# Package meta-data.
NAME = 'whatsonpypi'
VERSION = '0.2.5'
DESCRIPTION = "CLI tool to get package info from PyPI and/or manipulate requirements."
AUTHOR = "Visesh Prasad"
EMAIL = 'viseshrprasad@gmail.com'
URL = 'https://github.com/viseshrp/whatsonpypi'
REQUIRES_PYTHON = ">=2.7"
REQUIREMENTS = ['future>=0.15.2', 'Click>=7.0', 'requests>=2.18.0', ]
SETUP_REQUIREMENTS = ['pytest-runner', ]
TEST_REQUIREMENTS = ['pytest', ]

with io.open('README.rst', 'r', encoding='utf-8') as readme_file:
    README = readme_file.read()

with io.open('HISTORY.rst', 'r', encoding='utf-8') as history_file:
    HISTORY = history_file.read()

setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=README + '\n\n' + HISTORY,
    keywords='whatsonpypi',
    author=AUTHOR,
    author_email=EMAIL,
    url=URL,
    packages=find_packages(include=['whatsonpypi']),
    include_package_data=True,
    license="MIT license",
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Natural Language :: English',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    entry_points={
        'console_scripts': [
            'whatsonpypi=whatsonpypi.__main__:main',
        ],
    },
    python_requires=REQUIRES_PYTHON,
    install_requires=REQUIREMENTS,
    setup_requires=SETUP_REQUIREMENTS,
    test_suite='tests',
    tests_require=TEST_REQUIREMENTS,
    zip_safe=False,
)



# -*- coding: utf-8 -*-

"""
API client
"""
from __future__ import unicode_literals  # unicode support for py2

from requests import Request, Session, hooks
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

from .constants import PYPI_BASE_URL
from .exceptions import PackageNotProvidedError, PackageNotFoundError


class WoppResponse(object):
    """
    Serializer for the response from PyPI
    """

    def __init__(self, status_code, json):
        self.status_code = status_code
        self.json = json
        self.ok = self.status_code < 400

    @property
    def name(self):
        return self.json.get('name')

    @property
    def latest_version(self):
        return self.json.get('latest_version')

    @property
    def summary(self):
        return self.json.get('summary')

    @property
    def homepage(self):
        return self.json.get('homepage')

    @property
    def package_url(self):
        return self.json.get('package_url')

    @property
    def project_urls(self):
        return self.json.get('project_urls')

    @property
    def project_docs(self):
        return self.project_urls.get('Documentation') or self.homepage

    @property
    def requires_python(self):
        return self.json.get('requires_python')

    @property
    def license(self):
        return self.json.get('license')

    @property
    def author(self):
        return self.json.get('author')

    @property
    def author_email(self):
        return self.json.get('author_email')

    @property
    def latest_release_url(self):
        return self.json.get('latest_release_url')

    @property
    def dependencies(self):
        return self.json.get('dependencies')

    @property
    def latest_pkg_urls(self):
        return self.json.get('latest_pkg_urls')

    @property
    def releases(self):
        # all releases
        return self.json.get('releases')

    @property
    def releases_pkg_info(self):
        # info of every release's package
        return self.json.get('releases_pkg_info')

    @property
    def latest_releases(self):
        # last 5 releases
        return self.releases[:5]


class WoppClient(object):
    """
    Client for accessing the PyPI API
    """

    def __init__(self, pool_connections=True, request_hooks=None):
        # client specific info goes here.
        self.base_url = PYPI_BASE_URL
        self.session = Session() if pool_connections else None
        # default_hooks() returns {'response': []}
        self.request_hooks = request_hooks or hooks.default_hooks()

    def request(self, package=None, version=None, timeout=3.1, max_retries=3):
        """
        Make a HTTP GET request with the provided params

        :param timeout: request timeout seconds
        :param max_retries: number of times to retry on failure
        :param package: name of the python package to search
        :param version: version of the python package to search
        :return: response serialized by WoppResponse object
        """
        url = self._build_url(package, version)
        req_kwargs = {
            'method': 'GET',
            'url': url,
            'hooks': self.request_hooks,
            'headers': {
                'Accept': 'application/json',
                'User-Agent': 'https://github.com/viseshrp/whatsonpypi',
            }
        }

        session = self.session or Session()
        # instantiate Request
        request = Request(**req_kwargs)
        # Applies session-level state such as cookies to your request
        prepared_request = session.prepare_request(request)

        # use the adapter for retries
        retries = Retry(
            total=max_retries,
            backoff_factor=0.1,
            status_forcelist=[500, 502, 503, 504],
        )
        adapter = HTTPAdapter(max_retries=retries)
        session.mount('http://', adapter)
        session.mount('https://', adapter)

        # and fire!
        response = session.send(
            prepared_request,
            timeout=timeout,
            allow_redirects=True,
        )

        if response.status_code == 404:
            raise PackageNotFoundError("Sorry, but that package/version couldn't be found on PyPI.")

        # serialize response
        wopp_response = WoppResponse(int(response.status_code), response.cleaned_json)
        return wopp_response

    def _build_url(self, package=None, version=None):
        """
        Builds the URL with the path params provided.

        :param package: name of package
        :param version: version of package
        :return: fully qualified URL
        """
        if package is None:
            raise PackageNotProvidedError('A package name is needed to proceed.')

        if version is not None:
            url = "{}/{}/{}/json".format(self.base_url, package, version)
        else:
            url = "{}/{}/json".format(self.base_url, package)

        return url
